import warnings
from itertools import chain
from typing import Dict, Iterator, List, Optional, Set

from maggma.core import Builder, Store
from maggma.stores import S3Store
from monty.json import MontyDecoder
from pymatgen.analysis.phase_diagram import PhaseDiagramError
from pymatgen.entries.computed_entries import ComputedStructureEntry

from emmet.builders.utils import HiddenPrints
from emmet.core.thermo import ThermoDoc, PhaseDiagramDoc
from emmet.core.utils import jsanitize


class ThermoBuilder(Builder):
    def __init__(
        self,
        thermo: Store,
        corrected_entries: Store,
        phase_diagram: Optional[Store] = None,
        query: Optional[Dict] = None,
        num_phase_diagram_eles: Optional[int] = None,
        chunk_size: int = 1000,
        **kwargs,
    ):
        """
        Calculates thermodynamic quantities for materials from phase
        diagram constructions

        Args:
            thermo (Store): Store of thermodynamic data such as formation
                energy and decomposition pathway
            corrected_entries (Store): Store of corrected entry data to use in thermo data and phase diagram
                construction. This is required and should be built with the CorrectedEntriesBuilder.
            phase_diagram (Store): Store of phase diagram data for each unique chemical system
            query (dict): dictionary to limit materials to be analyzed
            num_phase_diagram_eles (int): Maximum number of elements to use in phase diagram construction
                for data within the separate phase_diagram collection
            chunk_size (int): Size of chemsys chunks to process at any one time.
        """

        self.thermo = thermo
        self.query = query if query else {}
        self.corrected_entries = corrected_entries
        self.phase_diagram = phase_diagram
        self.num_phase_diagram_eles = num_phase_diagram_eles
        self.chunk_size = chunk_size
        self._completed_tasks: Set[str] = set()

        if self.thermo.key != "thermo_id":
            warnings.warn(
                f"Key for the thermo store is incorrect and has been changed from {self.thermo.key} to thermo_id!"
            )
            self.thermo.key = "thermo_id"

        if self.corrected_entries.key != "chemsys":
            warnings.warn(
                "Key for the corrected entries store is incorrect and has been changed "
                f"from {self.corrected_entries.key} to chemsys!"
            )
            self.corrected_entries.key = "chemsys"

        sources = [corrected_entries]
        targets = [thermo]

        if self.phase_diagram is not None:

            if self.phase_diagram.key != "phase_diagram_id":
                warnings.warn(
                    f"Key for the phase diagram store is incorrect and has been changed from {self.phase_diagram.key} to phase_diagram_id!"  # noqa: E501
                )
                self.phase_diagram.key = "phase_diagram_id"

            targets.append(phase_diagram)  # type: ignore

        super().__init__(sources=sources, targets=targets, chunk_size=chunk_size, **kwargs)

    def ensure_indexes(self):
        """
        Ensures indicies on the tasks and materials collections
        """

        # Search index for corrected_entries
        self.corrected_entries.ensure_index("chemsys")
        self.corrected_entries.ensure_index("last_updated")

        # Search index for thermo
        self.thermo.ensure_index("material_id")
        self.thermo.ensure_index("thermo_id")
        self.thermo.ensure_index("thermo_type")
        self.thermo.ensure_index("last_updated")

        # Search index for thermo
        self.thermo.ensure_index("material_id")
        self.thermo.ensure_index("thermo_id")
        self.thermo.ensure_index("thermo_type")
        self.thermo.ensure_index("last_updated")

        # Search index for phase_diagram
        if self.phase_diagram:
            coll = self.phase_diagram

            if isinstance(self.phase_diagram, S3Store):
                coll = self.phase_diagram.index

            coll.ensure_index("chemsys")
            coll.ensure_index("phase_diagram_id")

    def get_items(self) -> Iterator[List[Dict]]:
        """
        Gets whole chemical systems of entries to process
        """

        self.logger.info("Thermo Builder Started")

        self.logger.info("Setting indexes")
        self.ensure_indexes()

        to_process_chemsys = self._get_chemsys_to_process()

        self.logger.info(f"Found {len(to_process_chemsys)} chemical systems with new/updated materials to process")
        self.total = len(to_process_chemsys)

        # Yield the chemical systems in order of increasing size
        # Will build them in a similar manner to fast Pourbaix
        for chemsys in sorted(to_process_chemsys, key=lambda x: len(x.split("-")), reverse=False):
            corrected_entries = self.corrected_entries.query_one({"chemsys": chemsys})
            yield corrected_entries

    def process_item(self, item):

        if not item:
            return None

        for thermo_type, entry_list in item["entries"].items():

            entries = [ComputedStructureEntry.from_dict(entry) for entry in entry_list]
            chemsys = item["chemsys"]
            elements = chemsys.split("-")

            self.logger.debug(f"Processing {len(entries)} entries for {chemsys} and thermo type {thermo_type}")

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                with HiddenPrints():

                    pd_thermo_doc_pair = self._produce_pair(entries, thermo_type, elements)

        return pd_thermo_doc_pair

    def _produce_pair(self, pd_entries, thermo_type, elements):
        # Produce thermo and phase diagram pair

        try:
            # Obtain phase diagram
            pd = ThermoDoc.construct_phase_diagram(pd_entries)

            # Iterate through entry material IDs and construct list of thermo docs to update
            docs = ThermoDoc.from_entries(pd_entries, thermo_type, pd, deprecated=False)

            pd_docs = [None]

            if self.phase_diagram:
                if self.num_phase_diagram_eles is None or len(elements) <= self.num_phase_diagram_eles:
                    chemsys = "-".join(sorted(set([e.symbol for e in pd.elements])))
                    pd_id = "{}_{}".format(chemsys, str(thermo_type))
                    pd_doc = PhaseDiagramDoc(
                        phase_diagram_id=pd_id, chemsys=chemsys, phase_diagram=pd, thermo_type=thermo_type,
                    )

                    pd_data = jsanitize(pd_doc.dict(), allow_bson=True)

                    pd_docs = [pd_data]

            docs_pd_pair = (
                jsanitize([d.dict() for d in docs], allow_bson=True),
                pd_docs,
            )

            return docs_pd_pair

        except PhaseDiagramError as p:
            elsyms = []
            for e in pd_entries:
                elsyms.extend([el.symbol for el in e.composition.elements])

            self.logger.error(f"Phase diagram error in chemsys {'-'.join(sorted(set(elsyms)))}: {p}")
            return (None, None)

    def update_targets(self, items):
        """
        Inserts the thermo and phase diagram docs into the thermo collection
        Args:
            items ([tuple(List[dict],List[dict])]): a list of thermo dictionaries to update
        """

        thermo_docs = [pair[0] for pair in items]
        phase_diagram_docs = [pair[1] for pair in items]

        # flatten out lists
        thermo_docs = list(filter(None, chain.from_iterable(thermo_docs)))
        phase_diagram_docs = list(filter(None, chain.from_iterable(phase_diagram_docs)))

        # Check if already updated this run
        thermo_docs = [i for i in thermo_docs if i["thermo_id"] not in self._completed_tasks]

        self._completed_tasks |= {i["thermo_id"] for i in thermo_docs}

        for item in thermo_docs:
            if isinstance(item["last_updated"], dict):
                item["last_updated"] = MontyDecoder().process_decoded(item["last_updated"])

        if self.phase_diagram:
            self.phase_diagram.update(phase_diagram_docs)

        if len(thermo_docs) > 0:
            self.logger.info(f"Updating {len(thermo_docs)} thermo documents")
            self.thermo.update(docs=thermo_docs, key=["thermo_id"])
        else:
            self.logger.info("No thermo items to update")

    def _get_chemsys_to_process(self):
        # Use last-updated to find new chemsys
        corrected_entries_chemsys_dates = {
            d[self.corrected_entries.key]: d[self.corrected_entries.last_updated_field]
            for d in self.corrected_entries.query(
                {}, properties=[self.corrected_entries.key, self.corrected_entries.last_updated_field]
            )
        }

        thermo_chemsys_dates = {}
        for d in self.thermo.query(
            {"deprecated": False}, properties=[self.corrected_entries.key, self.thermo.last_updated_field]
        ):

            entry = thermo_chemsys_dates.get(d[self.corrected_entries.key], None)
            if entry is None or d[self.thermo.last_updated_field] < entry:
                thermo_chemsys_dates[d[self.corrected_entries.key]] = d[self.thermo.last_updated_field]

        to_process_chemsys = [
            chemsys
            for chemsys in corrected_entries_chemsys_dates
            if (chemsys not in thermo_chemsys_dates)
            or (thermo_chemsys_dates[chemsys] < corrected_entries_chemsys_dates[chemsys])
        ]

        return to_process_chemsys



from maggma.builders.map_builder import MapBuilder
from maggma.core import Store
from pymatgen.core import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from emmet.core.bonds import BondingDoc
from emmet.core.utils import jsanitize


class BondingBuilder(MapBuilder):
    def __init__(
        self,
        oxidation_states: Store,
        bonding: Store,
        **kwargs,
    ):
        """
        Creates Bonding documents from structures, ideally with
        oxidation states already annotated but will also work from any
        collection with structure and mp-id.

        Args:
            oxidation_states: Store of oxidation states
            bonding: Store to update with bonding documents
            query : query on materials to limit search
        """
        self.oxidation_states = oxidation_states
        self.bonding = bonding
        self.kwargs = kwargs

        # Enforce that we key on material_id
        self.oxidation_states.key = "material_id"
        self.bonding.key = "material_id"
        super().__init__(
            source=oxidation_states,
            target=bonding,
            projection=["structure", "deprecated"],
            **kwargs,
        )

    def unary_function(self, item):
        structure = Structure.from_dict(item["structure"])
        mpid = item["material_id"]
        deprecated = item["deprecated"]

        # temporarily convert to conventional structure inside this builder,
        # in future do structure setting operations in a separate builder
        structure = SpacegroupAnalyzer(structure).get_conventional_standard_structure()

        bonding_doc = BondingDoc.from_structure(structure=structure, material_id=mpid, deprecated=deprecated)
        doc = jsanitize(bonding_doc.dict(), allow_bson=True)

        return doc



import json
import datetime
import copy

import pytest

from monty.io import zopen
from monty.serialization import loadfn

from maggma.stores import JSONStore, MemoryStore

from emmet.builders.qchem.molecules import MoleculesAssociationBuilder, MoleculesBuilder
from emmet.builders.molecules.atomic import PartialChargesBuilder, PartialSpinsBuilder
from emmet.builders.molecules.bonds import BondingBuilder
from emmet.builders.molecules.orbitals import OrbitalBuilder
from emmet.builders.molecules.redox import RedoxBuilder
from emmet.builders.molecules.thermo import ThermoBuilder
from emmet.builders.molecules.vibration import VibrationBuilder
from emmet.builders.molecules.summary import SummaryBuilder


@pytest.fixture(scope="session")
def tasks(test_dir):
    return JSONStore(test_dir / "builder_task_set.json.gz")


@pytest.fixture(scope="session")
def mols(tasks):
    assoc_store = MemoryStore(key="molecule_id")
    stage_one = MoleculesAssociationBuilder(tasks=tasks, assoc=assoc_store)
    stage_one.run()

    mol_store = MemoryStore(key="molecule_id")
    stage_two = MoleculesBuilder(assoc=assoc_store, molecules=mol_store, prefix="libe")
    stage_two.run()

    return mol_store


@pytest.fixture(scope="session")
def charges(test_dir):
    return MemoryStore(key="molecule_id")


@pytest.fixture(scope="session")
def spins(test_dir):
    return MemoryStore(key="molecule_id")


@pytest.fixture(scope="session")
def bonds(test_dir):
    return MemoryStore(key="molecule_id")


@pytest.fixture(scope="session")
def orbitals(test_dir):
    return MemoryStore(key="molecule_id")


@pytest.fixture(scope="session")
def redox(test_dir):
    return MemoryStore(key="molecule_id")


@pytest.fixture(scope="session")
def thermo(test_dir):
    return MemoryStore(key="molecule_id")


@pytest.fixture(scope="session")
def vibes(test_dir):
    return MemoryStore(key="molecule_id")


@pytest.fixture(scope="session")
def summary():
    return MemoryStore(key="molecule_id")


@pytest.mark.skip(reason="Waiting on molecule update.")
def test_summary_doc(tasks, mols, charges, spins, bonds, orbitals, redox, thermo, vibes, summary):
    charge_build = PartialChargesBuilder(tasks, mols, charges)
    charge_build.run()

    spins_build = PartialSpinsBuilder(tasks, mols, spins)
    spins_build.run()

    bonds_build = BondingBuilder(tasks, mols, bonds)
    bonds_build.run()

    orb_build = OrbitalBuilder(tasks, mols, orbitals)
    orb_build.run()

    redox_build = RedoxBuilder(mols, redox)
    redox_build.run()

    thermo_build = ThermoBuilder(tasks, mols, thermo)
    thermo_build.run()

    vibe_build = VibrationBuilder(tasks, mols, vibes)
    vibe_build.run()

    builder = SummaryBuilder(
        molecules=mols,
        charges=charges,
        spins=spins,
        bonds=bonds,
        orbitals=orbitals,
        redox=redox,
        thermo=thermo,
        vibes=vibes,
        summary=summary,
    )
    builder.run()

    docs = list(summary.query())

    assert len(docs) == 46



from pathlib import Path

import pytest
from maggma.stores import JSONStore, MemoryStore
from monty.serialization import dumpfn, loadfn

from emmet.builders.materials.corrected_entries import CorrectedEntriesBuilder
from emmet.builders.materials.thermo import ThermoBuilder
from emmet.builders.vasp.materials import MaterialsBuilder


@pytest.fixture(scope="session")
def tasks_store(test_dir):
    return JSONStore(test_dir / "test_si_tasks.json.gz")


@pytest.fixture(scope="session")
def materials_store(tasks_store):
    materials_store = MemoryStore(key="material_id")
    builder = MaterialsBuilder(tasks=tasks_store, materials=materials_store)
    builder.run()
    return materials_store


@pytest.fixture(scope="session")
def corrected_entries_store():
    return MemoryStore(key="chemsys")


@pytest.fixture
def thermo_store():
    return MemoryStore(key="thermo_id")


@pytest.fixture
def phase_diagram_store():
    return MemoryStore(key="chemsys")


def test_corrected_entries_builder(corrected_entries_store, materials_store):

    builder = CorrectedEntriesBuilder(materials=materials_store, corrected_entries=corrected_entries_store)
    builder.run()

    assert corrected_entries_store.count() == 1
    assert corrected_entries_store.count({"chemsys": "Si"}) == 1


def test_corrected_entries_serialization(tmpdir):
    builder = CorrectedEntriesBuilder(MemoryStore(), MemoryStore(), MemoryStore())

    dumpfn(builder.as_dict(), Path(tmpdir) / "test.json")
    loadfn(Path(tmpdir) / "test.json")


def test_thermo_builder(corrected_entries_store, thermo_store, phase_diagram_store):

    builder = ThermoBuilder(
        thermo=thermo_store, corrected_entries=corrected_entries_store, phase_diagram=phase_diagram_store
    )
    builder.run()

    assert thermo_store.count() == 1
    assert thermo_store.count({"material_id": "mp-149"}) == 1

    assert phase_diagram_store.count() == 1


def test_thermo_serialization(tmpdir):
    builder = ThermoBuilder(MemoryStore(), MemoryStore(), MemoryStore())

    dumpfn(builder.as_dict(), Path(tmpdir) / "test.json")
    loadfn(Path(tmpdir) / "test.json")



import pytest
from maggma.stores import JSONStore, MemoryStore

from emmet.builders.materials.basic_descriptors import BasicDescriptorsBuilder
from emmet.builders.vasp.materials import MaterialsBuilder
from emmet.builders.materials.similarity import StructureSimilarityBuilder


@pytest.fixture(scope="session")
def tasks_store(test_dir):
    return JSONStore(
        [test_dir / "test_si_tasks.json.gz", test_dir / "test_As2SO6_tasks.json.gz"]
    )


@pytest.fixture(scope="session")
def materials_store(tasks_store):
    materials_store = MemoryStore(key="material_id")
    builder = MaterialsBuilder(tasks=tasks_store, materials=materials_store)
    builder.run()
    return materials_store


@pytest.fixture(scope="session")
def descriptors_store(materials_store):
    descriptors_store = MemoryStore(key="task_id")
    builder = BasicDescriptorsBuilder(
        materials=materials_store, descriptors=descriptors_store
    )
    builder.run()
    return descriptors_store


def test_basic_descriptions(descriptors_store):
    similarity_store = MemoryStore()
    builder = StructureSimilarityBuilder(
        structure_similarity=similarity_store, site_descriptors=descriptors_store
    )
    builder.run()

    print(similarity_store.query_one({}))
    assert similarity_store.count() == 1
